@@ -44,13 +44,15 @@ mpi.scatterv <- function(x, scounts, type, rdata, root=0, comm=1){
 
 mpi.scatter.Robj <- function(obj=NULL, root=0, comm=1){
     if (mpi.comm.rank(comm) == root){
-     size<-mpi.comm.size(comm)
-        subobj<-lapply(obj,serialize, connection=NULL)
-    sublen<-unlist(lapply(subobj,length))
+		size<-mpi.comm.size(comm)
+        #subobj<-lapply(obj,serialize, connection=NULL)
+		subobj<-lapply(1:size, function(i) serialize(obj[[i]], NULL))
+
+		sublen<-unlist(lapply(subobj,length))
         #newsubobj<-strings.link(subobj,string(sum(sublen)+1))
-    newsubobj<-c(subobj,recursive=TRUE)
+		newsubobj<-c(subobj,recursive=TRUE)
         strnum<-mpi.scatter(sublen,type=1,rdata=integer(1),root=root,comm=comm)
-    outobj<-unserialize(mpi.scatterv(newsubobj,scounts=sublen,type= 4,
+		outobj<-unserialize(mpi.scatterv(newsubobj,scounts=sublen,type= 4,
                 rdata=raw(strnum),root=root, comm=comm))
     }
     else {
@@ -58,6 +60,7 @@ mpi.scatter.Robj <- function(obj=NULL, root=0, comm=1){
         outobj<-unserialize(mpi.scatterv(raw(strnum),scounts=0, type=4,
                 rdata=raw(strnum), root=root, comm=comm))
     }
+	gc()
     return(outobj)
 }
 
@@ -71,8 +74,7 @@ mpi.scatter.Robj2slave=function (obj, comm = 1) {
     mpi.bcast.cmd(cmd = .tmpRobj <- mpi.scatter.Robj(comm = 1), 
         rank = 0, comm = comm)
     mpi.scatter.Robj(obj=c(list("master"),obj), root = 0, comm = comm)
-    mpi.bcast.cmd(cmd = assign(.tmpname$objname, .tmpRobj, 
-        envir = .GlobalEnv), rank = 0, comm = comm)
+    mpi.bcast.cmd(cmd = assign(.tmpname$objname, .tmpRobj), rank = 0, comm = comm)
 }
 
 mpi.gather.Robj <- function(obj=NULL, root=0, comm=1, ...){
@@ -88,11 +90,15 @@ mpi.gather.Robj <- function(obj=NULL, root=0, comm=1, ...){
     cutobj=list()
     for(i in 1:size)
         cutobj[[i]]=allbiobj[(pos[i]+1):pos[i+1]]
-         sapply(cutobj,unserialize, ...)
+		out <- sapply(cutobj,unserialize, ...)
+		gc()
+		out
     }
     else {
          mpi.gather(bilen,type=1,rdata=integer(1),root=root,comm=comm)
-         mpi.gatherv(biobj,type=4,rdata=raw(1),rcounts=0,root=root,comm=comm)
+         out <- mpi.gatherv(biobj,type=4,rdata=raw(1),rcounts=0,root=root,comm=comm)
+		 gc()
+		 out
    }
 }
 
@@ -107,8 +113,9 @@ mpi.allgather.Robj <- function(obj=NULL, comm=1){
     cutobj=list()
     for(i in 1:size)
           cutobj[[i]]=allbiobj[(pos[i]+1):pos[i+1]]
-    sapply(cutobj,unserialize)
-
+    out <- sapply(cutobj,unserialize)
+	gc()
+	out
    # bistrcut<-sapply(rcounts,string)
    # bistr<-string.cut(allbiobj,bistrcut)
    # lapply(bistr,unserialize)
@@ -126,9 +133,9 @@ mpi.allgatherv <- function(x, type, rdata, rcounts, comm=1){
     as.integer(rcounts), as.integer(comm),PACKAGE = "Rmpi")
 }
 
-mpi.bcast <- function (x, type, rank = 0, comm = 1) {
+mpi.bcast <- function (x, type, rank = 0, comm = 1, buffunit=100) {
     .Call("mpi_bcast", .force.type(x,type), as.integer(type), as.integer(rank), 
-        as.integer(comm),PACKAGE = "Rmpi")
+        as.integer(comm), as.integer(buffunit), PACKAGE = "Rmpi")
 }
 
 #bin.nchar <- function(x){
@@ -137,17 +144,23 @@ mpi.bcast <- function (x, type, rank = 0, comm = 1) {
 #    .Call("bin_nchar", x[1],PACKAGE = "Rmpi")
 #}
 
-mpi.bcast.cmd <- function (cmd=NULL, rank=0, comm=1, nonblock=FALSE, sleep=0.1){
+mpi.bcast.cmd <- function (cmd=NULL, ..., rank=0, comm=1, nonblock=FALSE, sleep=0.1){
 	myrank=mpi.comm.rank(comm)
     if(myrank == rank){
-        cmd <- deparse(substitute(cmd), width.cutoff=500)
-		cmd <- serialize(cmd, NULL)
+        #cmd <- deparse(substitute(cmd), width.cutoff=500)
+		#cmd <- serialize(cmd, NULL)
+
+		scmd <- substitute(cmd)
+		arg <-list(...)
 		commsize <- mpi.comm.size(comm=comm)
+
+		scmd.arg <-serialize(list(scmd=scmd, arg=arg), NULL)
 		#mpi.bcast(x=length(cmd), type=1, rank=rank, comm=comm)
 		#invisible(mpi.bcast(x=cmd, type=4, rank=rank, comm=comm))
+		
 		for (i in 0:commsize) {
 			if (i != rank)
-				invisible(mpi.send(x=cmd, type=4, dest=i, tag=i, comm=comm))
+				invisible(mpi.send(x=scmd.arg, type=4, dest=i, tag=50000+i, comm=comm))
 			}
     } 
     else {
@@ -159,22 +172,29 @@ mpi.bcast.cmd <- function (cmd=NULL, rank=0, comm=1, nonblock=FALSE, sleep=0.1){
         #parse(text=out) 
         #}
 		if (!nonblock){
-			mpi.probe(mpi.any.source(), tag=myrank, comm)
+			mpi.probe(mpi.any.source(), tag=50000+myrank, comm)
 			srctag <- mpi.get.sourcetag(0)
 			charlen <- mpi.get.count(type=4, 0)
-			out <- unserialize(mpi.recv(x=raw(charlen), type=4,srctag[1],myrank, comm))
+			#out <- unserialize(mpi.recv(x=raw(charlen), type=4,srctag[1],srctag[2], comm))
+			scmd.arg <- unserialize(mpi.recv(x=raw(charlen), type=4,srctag[1],srctag[2], comm))
+
 		} else {
 			repeat {
-				if (mpi.iprobe(mpi.any.source(),tag=myrank,comm)){ 
+				if (mpi.iprobe(mpi.any.source(),tag=50000+myrank,comm)){ 
 					srctag <- mpi.get.sourcetag()
 					charlen <- mpi.get.count(type=4)
-					out <- unserialize(mpi.recv(x = raw(charlen), type = 4, srctag[1],myrank, comm))
+					#out <- unserialize(mpi.recv(x = raw(charlen), type = 4, srctag[1],srctag[2], comm))
+					scmd.arg <- unserialize(mpi.recv(x = raw(charlen), type = 4, srctag[1],srctag[2], comm))
 					break
 				}
 				Sys.sleep(sleep)
 			}
 		}			
-		parse(text=out)
+		#parse(text=out)
+		if (length(scmd.arg$arg)>0)
+			enquote(do.call(as.character(scmd.arg$scmd), scmd.arg$arg, envir=.GlobalEnv))
+		else 
+			scmd.arg$scmd
 	}
 }
 
@@ -182,23 +202,96 @@ mpi.bcast.Robj <- function(obj=NULL, rank=0, comm=1){
     if (mpi.comm.rank(comm) == rank){
     tmp <- serialize(obj, NULL)
     mpi.bcast(as.integer(length(tmp)), 1, rank, comm)
-    invisible(mpi.bcast(tmp, 4, rank, comm))
+    mpi.bcast(tmp, 4, rank, comm)
+	invisible(gc())
     }
     else {
     charlen <- mpi.bcast(integer(1), 1, rank, comm)
-    unserialize(mpi.bcast(raw(charlen), 4, rank, comm))
+    out <- unserialize(mpi.bcast(raw(charlen), 4, rank, comm))
+	gc()
+	out
     }
 }
 
-mpi.bcast.Robj2slave <- function(obj, comm=1){
-    objname <- deparse(substitute(obj),width.cutoff=500)
-    obj <- list(objname=objname,obj=obj)
-    mpi.bcast.cmd(cmd=.tmpRobj <- mpi.bcast.Robj(comm=1),
+mpi.bcast.Robj2slave <- function(obj, comm=1, all=FALSE){
+    if (!all){
+		objname <- deparse(substitute(obj),width.cutoff=500)
+		obj <- list(objname=objname,obj=obj)
+		mpi.bcast.cmd(cmd=.tmpRobj <- mpi.bcast.Robj(comm=comm),
                     rank=0, comm=comm)
-    mpi.bcast.Robj(obj, rank=0, comm=comm)
-    mpi.bcast.cmd(cmd=assign(.tmpRobj$objname,.tmpRobj$obj, 
-            envir = .GlobalEnv), rank=0, comm=comm)
-    #mpi.bcast.cmd(rm(.tmpRobj,envir = .GlobalEnv), rank=0, comm=comm) 
+		mpi.bcast.Robj(obj, rank=0, comm=comm)
+		mpi.bcast.cmd(cmd=assign(.tmpRobj$objname,.tmpRobj$obj), rank=0, comm=comm)
+		#mpi.bcast.cmd(rm(.tmpRobj,envir = .GlobalEnv), rank=0, comm=comm) 
+	}
+	else {
+		master.objects <-objects(envir=.GlobalEnv)
+		obj.num=length(master.objects)
+		if (obj.num)
+			for (i in 1:obj.num){
+				mpi.bcast.cmd(cmd=.tmpRobj <- mpi.bcast.Robj(comm=1),
+                    rank=0, comm=comm)
+				mpi.bcast.Robj(list(objname=master.objects[i], obj=get(master.objects[i])), 
+					rank=0, comm=comm)
+				mpi.bcast.cmd(cmd=assign(.tmpRobj$objname,.tmpRobj$obj), rank=0, comm=comm)
+			}
+	}
+}
+
+mpi.bcast.Rfun2slave <- function(comm=1){
+	master.fun <-objects(envir=.GlobalEnv)
+	sync.index <- which(lapply(lapply(master.fun, get), is.function)==1)
+	obj.num=length(sync.index)
+	if (obj.num)
+		for (i in sync.index){
+			mpi.bcast.cmd(cmd=.tmpRobj <- mpi.bcast.Robj(comm=1),
+                   rank=0, comm=comm)
+			mpi.bcast.Robj(list(objname=master.fun[i], obj=get(master.fun[i])), 
+				rank=0, comm=comm)
+			mpi.bcast.cmd(cmd=assign(.tmpRobj$objname,.tmpRobj$obj), rank=0, comm=comm)
+		}
+}
+
+mpi.bcast.data2slave <- function(obj, comm=1, buffunit=100){
+	if (!is.numeric(obj) || (!is.vector(obj) && !is.matrix(obj)))
+		return (mpi.bcast.Robj2slave(obj, comm=comm))
+		#stop ("Please use mpi.bcast.Robj2slave")
+		
+	objname <- serialize(deparse(substitute(obj),width.cutoff=500),NULL)
+	obj.info = integer(4)
+	obj.info[1]=length(objname)
+	if (is.vector(obj)){
+		if (buffunit < 1 || buffunit >=2^31)
+			stop("buffunit must be an integer between 1 and 2^31-1")
+		obj.info[2]=buffunit
+		obj.info[3]=length(obj)%/%buffunit
+		obj.info[4]=length(obj)%%buffunit
+	}
+	if (is.matrix(obj)){
+		obj.dim <-dim(obj)
+		obj.info[2]=obj.dim[1]
+		obj.info[3]=obj.dim[2]
+		obj.info[4]=0
+	}
+	
+	mpi.bcast.cmd(.tinfo <- mpi.bcast(integer(4),type=1),rank=0,comm=comm)
+	mpi.bcast(obj.info,type=1,rank=0,comm=comm)
+	
+	mpi.bcast.cmd(.tname<-unserialize(mpi.bcast(raw(.tinfo[1]),type=4)), rank=0, comm=comm)
+	mpi.bcast(objname, type=4, rank=0, comm=comm)
+	
+	if (is.vector(obj)){
+		mpi.bcast.cmd(.tmp.obj <- mpi.bcast(double(.tinfo[2]*(.tinfo[3]+(.tinfo[4]>0))),type=5, buffunit=.tinfo[2]),rank=0,comm=comm)
+		mpi.bcast(obj,type=5,rank=0,comm=comm,buffunit=buffunit)
+		mpi.bcast.cmd(assign(.tname,.tmp.obj[1:(.tinfo[2]*.tinfo[3]+.tinfo[4])]), rank=0, comm=comm)
+		mpi.bcast.cmd(rm(".tmp.obj"))
+		mpi.bcast.cmd(gc())
+	}
+	if (is.matrix(obj)){
+		mpi.bcast.cmd(.tmp.obj <- mpi.bcast(matrix(double(.tinfo[2]*.tinfo[3]), nrow=.tinfo[2]),type=5, buffunit=.tinfo[2]),rank=0,comm=comm)
+		mpi.bcast(obj,type=5,rank=0,comm=comm,buffunit=obj.info[2])	
+		mpi.bcast.cmd(assign(.tname,.tmp.obj), rank=0, comm=comm)
+		mpi.bcast.cmd(gc())
+	}
 }
 
 mpi.send <- function (x, type,  dest, tag, comm=1){
@@ -214,13 +307,16 @@ mpi.recv <- function (x, type, source, tag, comm=1, status=0){
 
 mpi.send.Robj <- function(obj, dest, tag, comm=1){
     mpi.send(x=serialize(obj, NULL), type=4, dest=dest, tag=tag, comm=comm)
+	invisible(gc())
 }
 
 mpi.recv.Robj <- function(source, tag, comm=1, status=0){
     mpi.probe(source, tag, comm, status)
     srctag <- mpi.get.sourcetag(status)
     charlen <- mpi.get.count(type=4, status)
-    unserialize(mpi.recv(x=raw(charlen), type=4,srctag[1],srctag[2], comm, status))
+    out<-unserialize(mpi.recv(x=raw(charlen), type=4,srctag[1],srctag[2], comm, status))
+	#gc()
+	out
 }
 
 mpi.reduce <- function(x, type=2, 
@@ -282,9 +378,11 @@ mpi.irecv <- function (x, type, source, tag, comm=1, request=0){
     PACKAGE = "Rmpi"))
 }
 
-mpi.isend.Robj <- function(obj, dest, tag, comm=1,request=0)
+mpi.isend.Robj <- function(obj, dest, tag, comm=1,request=0){
     mpi.isend(x=serialize(obj, NULL), type=4, dest=dest, tag=tag, 
         comm=comm,request=request)
+	#invisible(gc())
+}
 
 mpi.wait <- function(request, status=0)
     invisible(.Call("mpi_wait",  as.integer(request), as.integer(status), PACKAGE = "Rmpi"))
@@ -376,7 +474,8 @@ mpi.request.maxsize <- function()
         as.integer(x),
         as.double(x),
         as.character(x),
-    as.raw(x))
+		as.raw(x),
+		as.double(x))
 }
 #.mpi.serialize<- function(obj){
 #    trans_obj=serialize(obj,NULL)
